// Generated from Flocc program
// see http://www.flocc.net/

#ifndef HEADER_H
#define HEADER_H

#include <cmath>
#include <mpi.h>

//#define DEBUG 1
#include "pointer.h"
#include "hashing.h"
#include "vecmaps.h"
#include "reparter.h"
#include "carttops.h"

// BASE TYPES 

typedef struct nullStruct {
  bool operator<(const nullStruct& b) const {
    return false;
  }
  bool operator==(const nullStruct& b) const {
    return true;
  }
} nullStruct;

// PRINTING

void printVal(int v) { printf("%i", v); }
void printVal(const char *v) { printf("%s", v); }
void printVal(nullStruct v) {
  printVal("null");
}
void printVal(int* v) {
  // only safe to print one, as don't know array length
  printVal(*v); 
} 

// MPI STUFF

// converts the rank of a process in one communicator to it's rank
// in another communicator.
void translateRank(MPI::Comm *srcComm, int srcRank, MPI::Comm *destComm, int *destRank) {
  MPI_Group orig_group, new_group; 
  orig_group = srcComm->Get_group();
  new_group = destComm->Get_group();
  MPI::Group::Translate_ranks(orig_group, 1, &srcRank, new_group, destRank);
} 

//typedef void MPI::Comm::Errhandler_fn(MPI::Comm &, int *, ... );
void throwErrHandler(MPI::Comm &comm, int *errCode, ...) {
  int len = 5000;
  char str[len];
  for (int i = 0; i < len; i++) str[i] = 0;
  MPI::Get_error_string(*errCode, str, len);
  std::cerr << "MPI Error: " << str << "\n";
  throw 1;
}

// TIME

#define MAX_TIME_HEADER 1024
int timeRank;
char timeHeader[MAX_TIME_HEADER];
double beginTime, middleTime;

void initTime() {
  // rank
  timeRank = MPI::COMM_WORLD.Get_rank();

  // start time
  beginTime = MPI_Wtime();
  middleTime = beginTime;
}

void logTime(const char* event, double *startTime) {
  if (timeRank == 0) {
    double End = MPI_Wtime();
    std::cerr << timeHeader << "\t" 
              << event << "\t" << (End - (*startTime)) << std::endl;
  }
  middleTime = MPI_Wtime();
}

#endif

// Global cartesian communicator
static const int ndims = 1;
static const int maxNodesPerDim = 4096;
static std::vector<int> dimSizes;
static MPI::Cartcomm cartComm;
static int coordRanks[maxNodesPerDim];
static int thisCoord[ndims];
static int thisRank, Nproc, rootRank;


 // user def reduction function
  void userFun191(const void* invec, void *inoutvec, int len, const MPI::Datatype& datatype) {
   const double* in = (const double*)invec;
   double* out = (double*)inoutvec;
   // TODO: WHEN SENDING PROPER DATATYPE, CHANGE (len / sizeof(double)) to len!
   for (int i = 0; i < (len / sizeof(double)); i++) {
     double val2 = out[i];
       out[i] = in[i] + val2;
  
  }
 }

MPI::Op v196;

 // user def reduction function
  void userFun211(const void* invec, void *inoutvec, int len, const MPI::Datatype& datatype) {
   const double* in = (const double*)invec;
   double* out = (double*)inoutvec;
   // TODO: WHEN SENDING PROPER DATATYPE, CHANGE (len / sizeof(double)) to len!
   for (int i = 0; i < (len / sizeof(double)); i++) {
     double val2 = out[i];
       out[i] = in[i] + val2;
  
  }
 }

MPI::Op v216;

double glob;

void run() {
// Init global cartesian topology
thisRank = MPI::COMM_WORLD.Get_rank();
Nproc = MPI::COMM_WORLD.Get_size();
dimSizes = flocc::findBestDims(Nproc, ndims);

bool periods[ndims]; for (int i = 0; i < ndims; i++) periods[i] = false;
cartComm = MPI::COMM_WORLD.Create_cart(ndims, &dimSizes.front(), periods, true);

{
 int coord[ndims];
for (coord[0] = 0; coord[0] < dimSizes[0]; coord[0]++) {
coordRanks[coord[0]] = cartComm.Get_cart_rank(coord);
}

}
rootRank = coordRanks[0];
v196.Init(&userFun191, true);

v216.Init(&userFun211, true);

  double v1;
  v1 = 0.0;
  double v2;
  v2 = 0.0;
  int v3;
  v3 = 400000000;
  double v217;
  v217 = 1.0;
// BEGIN genFloatAList
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN genFloatAList\n";
#endif
#ifdef DEBUG
std::cerr << "WARNING: genFloatAList ignores par dims and distributions over ALL nodes.\n";
#endif

    int v4 = v3 / Nproc + (thisRank < (v3%Nproc) ? 1 : 0);
    double* v7((double*) malloc(v4 * sizeof(double)));
    assert(v7 != NULL);  
    for (int v5 = 0; v5 < v4; v5++) {
      v7[v5] = v5;
    }
    int v9 = 0;  

    double v12 = 0.0;
    for (int v8 = 0; v8 < v4; v8++) {
      v12 += v7[v8];
    }

    double v11 = 0.0;
    glob = v12;  
    //if (true) {
      MPI_Allreduce(&glob, &v11, sizeof(double), MPI_PACKED, v196, MPI::COMM_WORLD);
    //}

// END readAList
#ifdef DEBUG
std::cerr << thisRank << ") END readAList\n";
#endif
// BEGIN countAList
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN countAList\n";
#endif
  int v197;
  v197 = v3;
// END countAList
#ifdef DEBUG
std::cerr << thisRank << ") END countAList\n";
#endif
  double v199;
  v199 = (double)v197;
  double v200;
  //v200 = v11 / v199;
  v200 = 1.2 / v199;

std::cout << "mean: " << v200 << "\n";

// BEGIN readAList
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN readAList\n";
#endif
int v202 = 0;
    // BEGIN reduceList init
        double v205 = 0.0;
        double v206 = 0.0;
      
    // END reduceList init
    
if (true) {
  for (int v201 = 0; v201 < v4; v201++) {
      // BEGIN mapList consume:
    double v203;
        double v229;
      v229 = v7[v201] - v200;
    v203 = v229 * v229;
    
    // BEGIN reduceList consume
      v206 += v203;
      
    // END reduceList consume
    
  // END mapList consume
  
}
v205 = v206;
}
    // BEGIN reduceList fin
    if (true) {
        v206 = v205; //(Lf:(Float )) = (Lf:(Float ))
      
      cartComm.Allreduce(&v206, &v205, sizeof(double), MPI_PACKED, v216);
    }
    if (cartComm != cartComm) {
      cartComm.Bcast(&v205, sizeof(double), MPI_PACKED, rootRank);
    }
    // END reduceList fin
    
// END readAList
#ifdef DEBUG
std::cerr << thisRank << ") END readAList\n";
#endif
  double v218;
  v218 = v199 - v217;
  double v219;
  v219 = v205 / v218;
  double v220;
  v220 = sqrt(v219);
// BEGIN printFloat
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN printFloat\n";
#endif
if (thisRank == rootRank) std::cout << "stddev: " << v220 << "\n";
// END printFloat
#ifdef DEBUG
std::cerr << thisRank << ") END printFloat\n";
#endif
// No stream called 'final' found in output streams.
}


int main(int argc, char** argv) {
  MPI::Init(argc, argv);
  int rank = MPI::COMM_WORLD.Get_rank();
  int nproc = MPI::COMM_WORLD.Get_size();
  double startTime = MPI::Wtime();

  run();

  double duration = MPI::Wtime() - startTime;
  if (rank == 0) {
    std::cerr << "prog\t" << nproc << "\t" << duration << "\n"; 
  }

  MPI::Finalize();
  return 0;
}
//



