// Modified by Tristan Aubrey-Jones 2014
// For comparison with Flocc generated version
// see http://www.flocc.net/

// https://github.com/wesleykendall/mpitutorial/blob/master/mpi_reduce_allreduce/stddev.c
// https://raw.githubusercontent.com/wesleykendall/mpitutorial/master/mpi_reduce_allreduce/stddev.c
// Author: Wes Kendall
// Copyright 2013 www.mpitutorial.com
// This code is provided freely with the tutorials on mpitutorial.com. Feel
// free to modify it for your own use. Any distribution of the code must
// either provide a link to www.mpitutorial.com or keep this header in tact.
//
// Program that computes the standard deviation of an array of elements in parallel using
// MPI_Reduce.
//
#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <math.h>
#include <assert.h>
#include <iostream>
  
#define N 400000000

using namespace std;

  // Creates an array of random numbers. Each number has a value from 0 - 1
    double *create_rand_nums(int num_elements) {
      double *rand_nums = (double *)malloc(sizeof(double) * num_elements);
      assert(rand_nums != NULL);
      int i;
      for (i = 0; i < num_elements; i++) {
        rand_nums[i] = i; //2.0; //(rand() / (float)RAND_MAX);
      }
      return rand_nums;
    }

int main(int argc, char** argv) {

  // Seed the random number generator to get different results each time
  srand(time(NULL));

  MPI_Init(NULL, NULL);

  double startTime = MPI_Wtime();

  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  int world_size;
  MPI_Comm_size(MPI_COMM_WORLD, &world_size);

  int num_elements_per_proc = (N / world_size); //atoi(argv[1]);
  num_elements_per_proc += (world_rank < (N % world_size) ? 1 : 0);
  std::cout << world_size << "\t" << world_rank << "\t" << num_elements_per_proc << "\n";

  // Create a random array of elements on all processes.
  srand(world_rank);
  double *rand_nums = NULL;
  rand_nums = create_rand_nums(num_elements_per_proc);

  // Sum the numbers locally
  double local_sum = 0;
  int i;
  for (i = 0; i < num_elements_per_proc; i++) {
    local_sum += rand_nums[i];
  }

  std::cerr << world_rank << ") total: " << local_sum << "\n";

  // Reduce all of the local sums into the global sum in order to
  // calculate the mean
  double global_sum;
  MPI_Allreduce(&local_sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM,
                MPI_COMM_WORLD);
  double mean = global_sum / (num_elements_per_proc * world_size);

  // Compute the local sum of the squared differences from the mean
  double local_sq_diff = 0;
  for (i = 0; i < num_elements_per_proc; i++) {
    local_sq_diff += (rand_nums[i] - mean) * (rand_nums[i] - mean);
  }

  // Reduce the global sum of the squared differences to the root process
  // and print off the answer
  double global_sq_diff;
  MPI_Reduce(&local_sq_diff, &global_sq_diff, 1, MPI_DOUBLE, MPI_SUM, 0,
             MPI_COMM_WORLD);

  // The standard deviation is the square root of the mean of the squared
  // differences.
  double duration = MPI_Wtime()-startTime;
  if (world_rank == 0) {
    double stddev = sqrt(global_sq_diff /
                        (num_elements_per_proc * world_size));
    printf("Mean - %f, Standard deviation = %f\n", mean, stddev);
  double duration = MPI::Wtime() - startTime;
    if (world_rank == 0) {
      std::cerr << "stddevm\t60000000\t" << world_size << "\t" << duration << "\n";
    }
  }

  // Clean up
  free(rand_nums);
 
  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Finalize();
}
